# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2016 Online SAS and Contributors. All Rights Reserved.
#                         Julien Castets <jcastets@scaleway.com>
#
# Licensed under the BSD 2-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the
# License at https://opensource.org/licenses/BSD-2-Clause

from . import API


class MetadataAPI(API):
    """ The metadata API is used to get info about a running instance.

    To authenticate the client, the API uses its IP address. The header
    X-Auth-Token is not needed.
    """

    base_url = 'http://169.254.42.42/'

    def __init__(self, **kwargs):

        assert 'auth_token' not in kwargs, \
            'auth_token is not required to query the metadata API'

        super(MetadataAPI, self).__init__(auth_token=None, **kwargs)

    def get_metadata(self, as_shell=False):
        """ Returns server metadata.

        If `as_shell` is True, return a string easily parsable by a shell. If
        False, return a dictionary.
        """
        if as_shell:
            return self.query().conf.get()
        return self.query().conf.get(format='json')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2016 Online SAS and Contributors. All Rights Reserved.
#                         Julien Castets <jcastets@scaleway.com>
#
# Licensed under the BSD 2-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the
# License at https://opensource.org/licenses/BSD-2-Clause

import logging

# To enable logging, the client application needs to configure the logging std
# module, by calling for instance logging.basicConfig(level=logging.INFO) or
# preferably logging.config.dictConfig.

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:  # pragma no cover
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

# Prevent message "No handlers could be found for logger "scaleway"" to be
# displayed.
logging.getLogger(__name__).addHandler(NullHandler())


__version__ = '1.8.1'



from . import API


class BillingAPI(API):

    base_url = 'https://billing.scaleway.com'

    def __init__(self, **kwargs):
        base_url = kwargs.pop('base_url', BillingAPI.base_url)
        super(BillingAPI, self).__init__(base_url=base_url, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2016 Online SAS and Contributors. All Rights Reserved.
#                         Julien Castets <jcastets@scaleway.com>
#                         Kevin Deldycke <kdeldycke@scaleway.com>
#
# Licensed under the BSD 2-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the
# License at https://opensource.org/licenses/BSD-2-Clause

from __future__ import print_function

import logging
import platform
import sys
import time

import requests
import slumber

from .. import __version__


logger = logging.getLogger(__name__)


class _CustomHTTPAdapter(requests.adapters.HTTPAdapter):
    """ In order to support SNI in Python 2.x, the packages pyOpenSSL, pyasn1
    and ndg-httpsclient need to be installed. pyOpenSSL needs the system
    packages gcc, python-dev, libffi-dev and libssl-dev to be installed.

    Because Python packaging sucks, you will succeed to install pyOpenSSL even
    if the system requirements aren't installed ; but SNI requests will fail.

    _CustomHTTPAdapter is a simple wrapper around a requests HTTPAdapter that
    logs an explicit message if a SSLError occurs, as there are good chances
    the problem comes from a bad installation.
    """
    def send(self, *args, **kwargs):  # pragma: no cover
        try:
            return super(_CustomHTTPAdapter, self).send(*args, **kwargs)
        except requests.exceptions.SSLError:
            print("SSL error is raised by python-requests. This is probably "
                  "because the required modules to handle SNI aren't "
                  "installed correctly. You should probably uninstall them "
                  "(pip uninstall pyopenssl pyasn1 ndg-httpsclient), install "
                  "the system dependencies required for their installation "
                  "(on Ubuntu, apt-get install python-dev libffi-dev "
                  "libssl-dev) and resintall them (pip install pyopenssl "
                  "pyasn1 ndg-httpsclient).", file=sys.stderr)
            raise


class SlumberResource(slumber.Resource):

    # Maximum number of times we try to make a request against an API in
    # maintenance before aborting.
    MAX_RETRIES = 3

    def retry_in(self, retry):
        """ If the API returns a maintenance HTTP status code, sleep a while
        before retrying.
        """
        return min(2 ** retry, 30)

    def _request(self, *args, **kwargs):
        """ Makes a request to the Scaleway API, and wait patiently if there is
        an ongoing maintenance.
        """
        retry = 0

        while True:
            try:
                return super(SlumberResource, self)._request(*args, **kwargs)
            except slumber.exceptions.HttpServerError as exc:
                # Not a maintenance exception
                if exc.response.status_code not in (502, 503, 504):
                    raise

                retry += 1
                retry_in = self.retry_in(retry)

                if retry >= self.MAX_RETRIES:
                    logger.error(
                        'API endpoint still in maintenance after %s attempts. '
                        'Stop trying.' % (self.MAX_RETRIES,)
                    )
                    raise

                logger.info(
                    'API endpoint is currently in maintenance. Try again in '
                    '%s seconds... (retry %s on %s)' % (
                        retry_in, retry, self.MAX_RETRIES
                    )
                )
                time.sleep(retry_in)

    def _process_response(self, resp):
        if self._store.get('serialize', True) is False:
            return resp
        return super(SlumberResource, self)._process_response(resp)


class SlumberAPI(slumber.API):

    resource_class = SlumberResource


class API(object):

    base_url = None
    user_agent = 'scw-sdk/%s Python/%s %s' % (
        __version__, ' '.join(sys.version.split()), platform.platform()
    )

    def __init__(self, auth_token=None, base_url=None, verify_ssl=True,
                 user_agent=None, auth_jwt=None):

        self.auth_token = auth_token
        self.auth_jwt = auth_jwt
        if user_agent is not None:
            self.user_agent = user_agent

        if base_url:
            self.base_url = base_url

        self.verify_ssl = verify_ssl

    def make_requests_session(self):
        """ Attaches headers needed to query Scaleway APIs.
        """
        session = requests.Session()

        session.headers.update({'User-Agent': self.user_agent})

        if self.auth_token:
            # HTTP headers must always be ISO-8859-1 encoded
            session.headers.update({
                'X-Auth-Token': self.auth_token.encode('latin1')
            })
        if self.auth_jwt:
            session.headers.update({
                'X-Session-Token': self.auth_jwt.encode('latin1')
            })
        if not self.verify_ssl:
            session.verify = False

        session.mount('https://', _CustomHTTPAdapter())

        return session

    def get_api_url(self):
        return self.base_url

    def query(self, serialize=True, **kwargs):
        """ Gets a configured slumber.API object.
        """
        api = SlumberAPI(
            self.get_api_url(),
            session=self.make_requests_session(),
            **kwargs
        )
        api._store['serialize'] = serialize
        return api


from .api_account import AccountAPI  # noqa  # isort:skip
from .api_compute import ComputeAPI  # noqa  # isort:skip
from .api_metadata import MetadataAPI  # noqa  # isort:skip
from .api_billing import BillingAPI  # noqa  # isort:skip



# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2016 Online SAS and Contributors. All Rights Reserved.
#                         Julien Castets <jcastets@scaleway.com>
#                         Kevin Deldycke <kdeldycke@scaleway.com>
#
# Licensed under the BSD 2-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the
# License at https://opensource.org/licenses/BSD-2-Clause

import json

import httpretty
from six.moves.urllib.parse import urljoin


class FakeAPITestCase(object):

    def setUp(self):
        httpretty.enable()

    def tearDown(self):
        httpretty.disable()
        httpretty.reset()

    def fake_endpoint(self, api, endpoint, method=httpretty.GET,
                      body=None, status=200):

        if not callable(body):
            body = json.dumps(body)

        httpretty.register_uri(
            method,
            urljoin(api.get_api_url(), endpoint),
            body=body,
            content_type='application/json',
            status=status
        )



# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2016 Online SAS and Contributors. All Rights Reserved.
#                         Julien Castets <jcastets@scaleway.com>
#                         Kevin Deldycke <kdeldycke@scaleway.com>
#
# Licensed under the BSD 2-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the
# License at https://opensource.org/licenses/BSD-2-Clause

from . import API

REGIONS = {
    'par1': {
        'url': 'https://cp-par1.scaleway.com/',
    },
    'ams1': {
        'url': 'https://cp-ams1.scaleway.com/',
    }
}


class ComputeAPI(API):
    """ The default region is par1 as it was the first availability zone
    provided by Scaleway, but it could change in the future.
    """

    def __init__(self, **kwargs):
        region = kwargs.pop('region', None)
        base_url = kwargs.pop('base_url', None)

        assert region is None or base_url is None, \
            "Specify either region or base_url, not both."

        if base_url is None:
            region = region or 'par1'

            assert region in REGIONS, \
                "'%s' is not a valid Scaleway region." % region

            base_url = REGIONS.get(region)['url']

        super(ComputeAPI, self).__init__(base_url=base_url, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2013-2016 Online SAS and Contributors. All Rights Reserved.
#                         Julien Castets <jcastets@scaleway.com>
#                         Kevin Deldycke <kdeldycke@scaleway.com>
#
# Licensed under the BSD 2-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the
# License at https://opensource.org/licenses/BSD-2-Clause

import unittest

import mock
import requests
import slumber
from scaleway.apis import API, SlumberResource

from . import FakeAPITestCase


class SimpleAPI(API):

    base_url = 'http://localhost'


class TestAPI(FakeAPITestCase, unittest.TestCase):

    def test_make_requests_session(self):
        # normal use
        requests_session = API(
            auth_token='0xdeadbeef', user_agent='jamesb0nd'
        ).make_requests_session()

        self.assertEqual(requests_session.headers.get('X-Auth-Token'),
                         b'0xdeadbeef')
        self.assertEqual(requests_session.headers.get('User-Agent'),
                         'jamesb0nd')

        # no auth provided
        session_of_no_auth_api = API().make_requests_session()
        self.assertIsInstance(session_of_no_auth_api, requests.Session)
        self.assertNotIn('X-Auth-Token', session_of_no_auth_api.headers)

        # check flag verify_ssl
        self.assertTrue(API().make_requests_session().verify)
        self.assertFalse(API(verify_ssl=False).make_requests_session().verify)

        # HTTP headers must be latin1 encoded
        token_unicode = u'éè'
        session = SimpleAPI(auth_token=token_unicode).make_requests_session()
        self.assertEqual(
            session.headers['X-Auth-Token'],
            token_unicode.encode('latin1')
        )

    def test_get_api_url(self):
        self.assertEqual(SimpleAPI().get_api_url(), 'http://localhost')

        self.assertEqual(
            SimpleAPI(base_url='http://hello').get_api_url(), 'http://hello'
        )

    @mock.patch('time.sleep', return_value=None)
    def test_maintenance(self, sleep):
        api = SimpleAPI()
        self.fake_endpoint(api, 'whatever/', status=503)

        self.assertRaises(
            slumber.exceptions.HttpServerError,
            api.query().whatever.get
        )
        self.assertEqual(sleep.call_count, SlumberResource.MAX_RETRIES - 1)
        sleep.assert_called_with(4)

    def test_append_slash(self):
        api = SimpleAPI()
        self.fake_endpoint(api, 'slash/', status=200, body='slash')
        self.fake_endpoint(api, 'no_slash', status=200, body='no slash')

        # Default is to set append_slash
        self.assertEqual(api.query().slash.get(), 'slash')
        self.assertEqual(api.query(append_slash=True).slash.get(), 'slash')

        self.assertEqual(
            api.query(append_slash=False).no_slash.get(), 'no slash'
        )

    def test_serialize(self):
        api = SimpleAPI()
        self.fake_endpoint(api, 'whatever/', status=200, body='xxx')

        response = api.query(serialize=False).whatever.get()
        self.assertIsInstance(response, requests.Response)

        response = api.query(serialize=True).whatever.get()
        try:
            self.assertIsInstance(response, unicode)
        except NameError:  # python3
            self.assertIsInstance(response, str)
